from pycsghub.snapshot_download import snapshot_download
from pycsghub.file_upload import http_upload_file
from pathlib import Path
from typing import Optional, Union, List
from pycsghub.constants import DEFAULT_REVISION
from pycsghub.repository import Repository
from pycsghub.utils import get_token_to_send

def download(
        repo_id: str,
        repo_type: str,
        revision: Optional[str] = DEFAULT_REVISION,
        cache_dir: Union[str, Path, None] = None,
        endpoint: Optional[str] = None,
        token: Optional[str] = None,
        allow_patterns: Optional[Union[List[str], str]] = None,
        ignore_patterns: Optional[Union[List[str], str]] = None,
        source: str = None,
    ):
    snapshot_download(
        repo_id=repo_id,
        repo_type=repo_type,
        revision=revision,
        cache_dir=cache_dir, 
        endpoint=endpoint, 
        token=token,
        allow_patterns=allow_patterns,
        ignore_patterns=ignore_patterns,
        source=source,
    )

def upload_files(
        repo_id: str,
        repo_type: str,
        repo_file: str,
        path_in_repo: Optional[str] = "",
        revision: Optional[str] = DEFAULT_REVISION,
        endpoint: Optional[str] = None,
        token: Optional[str] = None
    ):
    http_upload_file(
        repo_id=repo_id,
        repo_type=repo_type,
        file_path=repo_file,
        path_in_repo=path_in_repo,
        revision=revision,
        endpoint=endpoint,
        token=token,
    )

def upload_folder(
        repo_id: str,
        repo_type: str,
        local_path: str,
        path_in_repo: Optional[str] = "",
        work_dir: Optional[str] = "/tmp/csg",
        nickname: Optional[str] = "",
        description: Optional[str] = "",
        license: Optional[str] = "apache-2.0",
        revision: Optional[str] = DEFAULT_REVISION,
        endpoint: Optional[str] = None,
        user_name: Optional[str] = "",
        token: Optional[str] = None,
        auto_create: Optional[bool] = True,
    ):
    r = Repository(
        repo_id=repo_id,
        upload_path=local_path,
        path_in_repo=path_in_repo,
        work_dir=work_dir,
        repo_type=repo_type,
        nickname=nickname,
        description=description,
        license=license,
        branch_name=revision,
        endpoint=endpoint,
        user_name=user_name,
        token=get_token_to_send(token),
        auto_create=auto_create,
    )
    r.upload()
    



import io
import requests
from tqdm import tqdm
import logging
from .status import JOB_ITEM_T
from typing import Dict

logger = logging.getLogger(__name__)

class UploadTracker(io.BytesIO):
    def __init__(self, data, progress_bar):
        super().__init__(data)
        self._progress_bar = progress_bar

    def read(self, size=-1):
        chunk = super().read(size)
        self._progress_bar.update(len(chunk))
        return chunk

def slice_upload(item: JOB_ITEM_T):
    paths, metadata = item
    upload_desc = f"uploading {paths.file_path}({metadata.lfs_upload_part_index}/{metadata.lfs_upload_part_count})"
    
    read_chunk_size = metadata.lfs_upload_chunk_size
    if metadata.lfs_upload_part_index == metadata.lfs_upload_part_count:
        read_chunk_size = metadata.size - (metadata.lfs_upload_part_count - 1) * metadata.lfs_upload_chunk_size
    
    chunk_data = None
    with paths.file_path.open('rb') as f:
        f.seek((metadata.lfs_upload_part_index - 1) * metadata.lfs_upload_chunk_size)
        chunk_data = f.read(read_chunk_size)

    total = len(chunk_data)
    headers = {
        "Content-Type": "application/octet-stream",
        "Content-Length": str(total)
    }
    with tqdm(initial=0, total=total, desc=upload_desc, unit="B", unit_scale=True, dynamic_ncols=True) as pbar:
        upload_data = UploadTracker(chunk_data, pbar)
        response = requests.put(
            url=metadata.lfs_upload_part_url,
            headers=headers,
            data=upload_data,
        )
        if response.status_code != 200:
            logger.error(f"LFS slice {paths.file_path}({metadata.lfs_upload_part_index}/{metadata.lfs_upload_part_count}) upload on {metadata.lfs_upload_part_url} response: {response.text}")
        response.raise_for_status()
        return response.headers

def slices_upload_complete(item: JOB_ITEM_T, uploaded_ids_map: Dict):
    paths, metadata = item
    payload = {
        "oid": metadata.sha256,
        "uploadId": metadata.lfs_upload_id,
        "parts": [
            {"partNumber": i+1, "etag": f"{uploaded_ids_map.get(str(i+1))}"}
            for i in range(metadata.lfs_upload_part_count)
        ]
    }
    response = requests.post(metadata.lfs_upload_complete_url, json=payload)
    if response.status_code != 200 and (response.status_code < 400 or response.status_code >= 500):
        logger.error(f"LFS {paths.file_path} merge all uploaded slices complete on {metadata.lfs_upload_complete_url} response: {response.text}")
    if response.status_code < 400 or response.status_code >= 500:
        response.raise_for_status()
    return response.text

def slices_upload_verify(item: JOB_ITEM_T):
    paths, metadata = item
    payload = {
        "oid": metadata.sha256,
        "size": metadata.size,
    }
    verify_url = metadata.lfs_upload_verify.get("href")
    verify_header = metadata.lfs_upload_verify.get("header")
    response = requests.post(verify_url, headers=verify_header, json=payload)
    if response.status_code != 200:
        logger.error(f"LFS {paths.file_path} slices uploaded verify on {verify_url} response: {response.text}, delete file {paths.metadata_path} and retry")
    response.raise_for_status()
    return response.text



import requests
from pycsghub.utils import (build_csg_headers,
                            get_endpoint)

def list(
    user_name: str,
    token: str,
    endpoint: str,
    limit: int,
):
    action_endpoint = get_endpoint(endpoint=endpoint)
    url = f"{action_endpoint}/api/v1/user/{user_name}/finetune/instances"
    data = {
        "per": limit,
    }
    headers = build_csg_headers(token=token, headers={
        "Content-Type": "application/json"
    })
    response = requests.get(url, params=data, headers=headers)
    response.raise_for_status()
    instances = response.json()['data']
    print(f"{'ID':<10}{'Name':<40}{'Model':<50}{'Status':<10}")
    print("-" * 110)
    if instances:
        for instance in instances:
            print(f"{instance['deploy_id']:<10}"
                f"{instance['deploy_name']:<40}"
                f"{instance['model_id']:<50}"
                f"{instance['status']:<10}")

def start(
    id: int,
    model: str,
    token: str,
    endpoint: str,
):
    action_endpoint = get_endpoint(endpoint=endpoint)
    url = f"{action_endpoint}/api/v1/models/{model}/finetune/{id}/start"
    headers = build_csg_headers(token=token, headers={
        "Content-Type": "application/json"
    })
    response = requests.put(url, headers=headers)
    result = response.json()
    print(result)

def stop(
    id: int,
    model: str,
    token: str,
    endpoint: str,
):
    action_endpoint = get_endpoint(endpoint=endpoint)
    url = f"{action_endpoint}/api/v1/models/{model}/finetune/{id}/stop"
    headers = build_csg_headers(token=token, headers={
        "Content-Type": "application/json"
    })
    response = requests.put(url, headers=headers)
    result = response.json()
    print(result)



import os
import logging
import threading
import time
from pathlib import Path
from tqdm.auto import tqdm
from typing import Optional, Union, List
from pycsghub.cmd.repo_types import RepoType
from pycsghub.constants import REPO_TYPE_MODEL, REPO_TYPE_DATASET, REPO_TYPE_SPACE
from pycsghub.utils import get_endpoint
from .path import filter_repo_objects
from .local_folder import get_local_upload_paths, read_upload_metadata
from .workers import _worker_job
from .status import LargeUploadStatus
from .consts import DEFAULT_IGNORE_PATTERNS
from pycsghub.csghub_api import CsgHubApi
from pycsghub.constants import DEFAULT_REVISION

logger = logging.getLogger(__name__)

def upload_large_folder_internal(
    repo_id: str,
    local_path: str,
    repo_type: RepoType,
    revision: str,
    endpoint: str,
    token: str,
    allow_patterns: Optional[Union[List[str], str]],
    ignore_patterns: Optional[Union[List[str], str]],
    num_workers: Optional[int],
    print_report: bool,
    print_report_every: int,
):
    folder_path = Path(local_path).expanduser().resolve()
    if not folder_path.is_dir():
        raise ValueError(f"provided path '{local_path}' is not a directory")
    
    if repo_type not in [REPO_TYPE_MODEL, REPO_TYPE_DATASET, REPO_TYPE_SPACE]:
        raise ValueError(f"invalid repo type, must be one of {REPO_TYPE_MODEL} or {REPO_TYPE_DATASET} or {REPO_TYPE_SPACE}")
    
    api_endpoint = get_endpoint(endpoint=endpoint)

    if ignore_patterns is None:
        ignore_patterns = []
    elif isinstance(ignore_patterns, str):
        ignore_patterns = [ignore_patterns]
    ignore_patterns += DEFAULT_IGNORE_PATTERNS
    
    if num_workers is None:
        nb_cores = os.cpu_count() or 1
        num_workers = max(nb_cores - 2, 2)
    
    api = CsgHubApi()
    
    create_repo(api=api, repo_id=repo_id, repo_type=repo_type, revision=revision, endpoint=api_endpoint, token=token)

    filtered_paths_list = filter_repo_objects(
        (path.relative_to(folder_path).as_posix() for path in folder_path.glob("**/*") if path.is_file()),
        allow_patterns=allow_patterns,
        ignore_patterns=ignore_patterns,
    )
    
    paths_list = [get_local_upload_paths(folder_path, relpath) for relpath in filtered_paths_list]
    
    items = [
        (paths, read_upload_metadata(folder_path, paths.path_in_repo))
        for paths in tqdm(paths_list, desc=f"recovering from cache metadata from {folder_path}/.cache")
    ]

    logger.info(f"starting {num_workers} worker threads for upload tasks")
    status = LargeUploadStatus(items)
    threads = [
        threading.Thread(
            target=_worker_job,
            kwargs={
                "status": status,
                "api": api,
                "repo_id": repo_id,
                "repo_type": repo_type,
                "revision": revision,
                "endpoint": api_endpoint,
                "token": token,
            },
        )
        for _ in range(num_workers)
    ]
    
    for thread in threads:
        thread.start()

    if print_report:
        print('\n' + status.current_report())
    last_report_ts = time.time()
    while True:
        time.sleep(1)
        if time.time() - last_report_ts >= print_report_every:
            if print_report:
                print(status.current_report())
            last_report_ts = time.time()
        if status.is_done():
            logging.info("all files are done and exiting main loop")
            break

    for thread in threads:
        thread.join()

    print(status.current_report())
    logging.info("large folder upload process is complete!")

def create_repo(
    api: CsgHubApi,
    repo_id: str,
    repo_type: str,
    revision: str,
    endpoint: str,
    token: str,
):
    repoExist, branchExist = api.repo_branch_exists(
        repo_id=repo_id, repo_type=repo_type, revision=revision,
        endpoint=endpoint, token=token)

    if not repoExist:
        api.create_new_repo(
            repo_id=repo_id, repo_type=repo_type, revision=revision,
            endpoint=endpoint, token=token)
        logger.info(f"repo {repo_type} {repo_id} created")
        if revision == DEFAULT_REVISION:
            branchExist = True

    if not branchExist:
        api.create_new_branch(
            repo_id=repo_id, repo_type=repo_type, revision=revision,
            endpoint=endpoint, token=token)



import unittest
from pycsghub.repo_reader import AutoModelForCausalLM
from pathlib import Path

class MyTestCase(unittest.TestCase):
    def test_something(self):
        self.assertEqual(True, False)  # add assertion here

    def test_something_else(self):
        a = AutoModelForCausalLM.from_pretrained('xzgan001/csg-wukong-1B')
        print(a)


if __name__ == '__main__':
    unittest.main()



import logging
import traceback
import time
import copy
from typing import Optional, List, Tuple, Dict
from .status import LargeUploadStatus, WorkerJob, JOB_ITEM_T
from .consts import WAITING_TIME_IF_NO_TASKS
from .jobs import _determine_next_job
from .sha import sha_fileobj
from pycsghub.csghub_api import CsgHubApi
from .utils import unquote
from .consts import (
    REPO_REGULAR_TYPE, 
    REPO_LFS_TYPE, 
    COMMIT_ACTION_CREATE, 
    COMMIT_ACTION_UPDATE,
    KEY_MSG, MSG_OK,
    KEY_UPLOADID
)
from urllib.parse import urlparse, parse_qs
from .slices import slice_upload, slices_upload_complete, slices_upload_verify

logger = logging.getLogger(__name__)

def _worker_job(
    status: LargeUploadStatus,
    api: CsgHubApi,
    repo_id: str,
    repo_type: str,
    revision: str,
    endpoint: str,
    token: str,
):
    """
    Main process for a worker. The worker will perform tasks based on the priority list until all files are uploaded
    and committed. If no tasks are available, the worker will wait for 10 seconds before checking again.

    If a task fails for any reason, the item(s) are put back in the queue for another worker to pick up.
    """
    while True:
        next_job: Optional[Tuple[WorkerJob, List[JOB_ITEM_T]]] = None

        # Determine next task
        next_job = _determine_next_job(status)
        if next_job is None:
            return
        job, items = next_job
        logger.debug(f"next job: {job}")
        # Perform task
        if job == WorkerJob.SHA256:
            _execute_job_compute_sha256(items=items, status=status)
        elif job == WorkerJob.GET_UPLOAD_MODE:
            _execute_job_get_upload_model(
                items=items, status=status,
                api=api, endpoint=endpoint, token=token,
                repo_id=repo_id, repo_type=repo_type, revision=revision)
        elif job == WorkerJob.PREUPLOAD_LFS:
            _execute_job_pre_upload_lfs(
                items=items, status=status,
                api=api, endpoint=endpoint, token=token,
                repo_id=repo_id, repo_type=repo_type, revision=revision)
        elif job == WorkerJob.UPLOADING_LFS:
            _execute_job_uploading_lfs(items=items, status=status)
        elif job == WorkerJob.COMMIT:
            _execute_job_commit(
                items=items, status=status,
                api=api, endpoint=endpoint, token=token,
                repo_id=repo_id, repo_type=repo_type, revision=revision)
        elif job == WorkerJob.WAIT:
            _execute_job_waiting(status=status)

def _execute_job_compute_sha256(
    items: List[JOB_ITEM_T], 
    status: LargeUploadStatus,
):
    item = items[0]  # single item every time
    paths, metadata = item
    try:
        _compute_sha256(item)
        logger.debug(f"computing sha256 for {item[0].file_path} successfully")
        status.queue_get_upload_mode.put(item)
    except KeyboardInterrupt:
        raise
    except Exception as e:
        logger.error(f"failed to compute {paths.file_path} sha256: {e}")
        traceback.format_exc()
        status.queue_sha256.put(item)

    with status.lock:
        status.nb_workers_sha256 -= 1

def _execute_job_get_upload_model(
    items: List[JOB_ITEM_T], 
    status: LargeUploadStatus,
    api: CsgHubApi,
    repo_id: str,
    repo_type: str,
    revision: str,
    endpoint: str,
    token: str,
):
    # A maximum of 50 files at a time
    try:
        _get_upload_mode(
            items, api=api, endpoint=endpoint, token=token,
            repo_id=repo_id, repo_type=repo_type, revision=revision)
        logger.debug(f"get upload modes for {len(items)} items successfully")
    except KeyboardInterrupt:
        raise
    except Exception as e:
        logger.error(f"failed to get {len(items)} items upload mode: {e}")
        traceback.format_exc()

    # Items are either:
    # - dropped (if should_ignore)
    # - put in LFS queue (if LFS)
    # - put in commit queue (if regular)
    # - or put back (if error occurred).
    ignore_num = 0
    same_with_remote_num = 0
    for item in items:
        paths, metadata = item
        if metadata.should_ignore:
            ignore_num += 1
            continue
        if ((metadata.upload_mode == REPO_REGULAR_TYPE and metadata.sha1 == metadata.remote_oid) or
            (metadata.upload_mode == REPO_LFS_TYPE and metadata.sha256 == metadata.remote_oid)):
            metadata.is_uploaded = True
            metadata.is_committed = True
            metadata.save(paths)
            same_with_remote_num += 1
            continue
        if metadata.upload_mode == REPO_LFS_TYPE:
            status.queue_preupload_lfs.put(item)
        elif metadata.upload_mode == REPO_REGULAR_TYPE:
            status.queue_commit.put(item)
        else:
            status.queue_get_upload_mode.put(item)

    if ignore_num > 0:
        logger.info(f"ignored {ignore_num} files because should_ignore is true from remote server")
    
    if same_with_remote_num > 0:
        logger.info(f"skipped {same_with_remote_num} files because they are identical to the remote server")
    
    with status.lock:
        status.nb_workers_get_upload_mode -= 1

def _execute_job_pre_upload_lfs(
    items: List[JOB_ITEM_T], 
    status: LargeUploadStatus,
    api: CsgHubApi,
    repo_id: str,
    repo_type: str,
    revision: str,
    endpoint: str,
    token: str,
):
    item = items[0]  # single item every time
    paths, metadata = item
    action = "preupload"
    try:
        if status.is_lfs_upload_completed(item):
            action = f"{action} check complete"
            _preupload_lfs_done(item=item, status=status)
            status.queue_commit.put(item)
        else:
            action = f"{action} fetch batch info"
            _preupload_lfs(
                item=item, status=status,
                api=api, endpoint=endpoint, token=token,
                repo_id=repo_id, repo_type=repo_type, revision=revision)
            # keep in queue preupload
            status.queue_preupload_lfs.put(item)
    except KeyboardInterrupt:
        raise
    except Exception as e:
        logger.error(f"failed to {action} lfs {paths.file_path}: {e}")
        traceback.format_exc()
        status.queue_preupload_lfs.put(item)

    with status.lock:
        status.nb_workers_preupload_lfs -= 1   

def _execute_job_uploading_lfs(
    items: List[JOB_ITEM_T], 
    status: LargeUploadStatus,
):
    item = items[0] # single item every time
    paths, metadata = item
    try:
        etag = _perform_lfs_slice_upload(item)
        status.append_lfs_uploaded_slice_id(paths.file_path, metadata.lfs_upload_part_index, etag)
        metadata.lfs_uploaded_ids = status.get_lfs_uploaded_slice_ids(paths.file_path)
        metadata.save(paths)
    except Exception as e:
        logger.error(f"failed to preupload LFS {paths.file_path} slice {metadata.lfs_upload_part_index}/{metadata.lfs_upload_part_count}: {e}")
        traceback.format_exc()
        status.queue_uploading_lfs.put(item)
        
    with status.lock:
        status.nb_workers_uploading_lfs -= 1

def _execute_job_commit(
    items: List[JOB_ITEM_T],
    status: LargeUploadStatus,
    api: CsgHubApi,
    repo_id: str,
    repo_type: str,
    revision: str,
    endpoint: str,
    token: str,
):
    try:
        for item in items:
            status.compute_file_base64(item=item)
        
        _commit(items, api=api, endpoint=endpoint, token=token,
            repo_id=repo_id, repo_type=repo_type, revision=revision)
        logger.info(f"committed {len(items)} items")
    except KeyboardInterrupt:
        raise
    except Exception as e:
        logger.error(f"failed to commit: {e}")
        traceback.format_exc()
        for item in items:
            status.queue_commit.put(item)

    with status.lock:
        status.last_commit_attempt = time.time()
        status.nb_workers_commit -= 1

def _execute_job_waiting(
    status: LargeUploadStatus,
):
    time.sleep(WAITING_TIME_IF_NO_TASKS)
    with status.lock:
        status.nb_workers_waiting -= 1    

def _compute_sha256(item: JOB_ITEM_T) -> None:
    """Compute sha256 of a file and save it in metadata."""
    paths, metadata = item
    if metadata.sha256 is None:
        with paths.file_path.open("rb") as f:
             sha256, sha1 = sha_fileobj(fileobj=f, item=item)
             metadata.sha256 = sha256
             metadata.sha1 = sha1
             
    metadata.save(paths)

def _get_upload_mode(
    items: List[JOB_ITEM_T],
    api: CsgHubApi, 
    repo_id: str, 
    repo_type: str, 
    revision: str,
    endpoint: str,
    token: str,
) -> None:
    """Get upload mode for each file and update metadata.

    Also receive info if the file should be ignored.
    """
    payload: Dict = {
        "files": [
            {
                "path": paths.path_in_repo,
                "size": meta.size,
            }
            for paths, meta in items
        ]
    }
    modes_resp = api.fetch_upload_modes(
        payload=payload, endpoint=endpoint, token=token,
        repo_id=repo_id, repo_type=repo_type, revision=revision)
    
    if modes_resp["data"] is None or modes_resp["data"]["files"] is None:
        raise ValueError("no correct upload modes response found")
        
    files_modes = modes_resp["data"]["files"]
    if not isinstance(files_modes, list):
        raise ValueError("files is not list in upload modes response")
    
    if not files_modes and len(files_modes) != len(items):
        raise ValueError(f"requested {len(items)} files do not match {len(files_modes)} files in fetch upload modes response")

    remote_upload_modes: Dict[str, str] = {}
    remote_should_ignore: Dict[str, bool] = {}
    remote_file_oids: Dict[str, Optional[str]] = {}
    
    for file in files_modes:
        key = file["path"]
        if file["isDir"]:
            raise ValueError(f"cannot upload '{key}' - the path exists as a directory in the remote repository")
        remote_upload_modes[key] = file["uploadMode"]
        remote_should_ignore[key] = file["shouldIgnore"]
        remote_file_oids[key] = file["oid"]
    
    for item in items:
        paths, metadata = item
        metadata.upload_mode = remote_upload_modes[paths.path_in_repo]
        metadata.should_ignore = remote_should_ignore[paths.path_in_repo]
        metadata.remote_oid = None if remote_file_oids[paths.path_in_repo] == "" else remote_file_oids[paths.path_in_repo]
        metadata.save(paths)

def _preupload_lfs_done(
    item: JOB_ITEM_T,
    status: LargeUploadStatus,
):
    paths, metadata = item
    uploaded_ids = status.get_lfs_uploaded_slice_ids(paths.file_path)
    uploaded_ids_map = status.convert_uploaded_ids_to_map(uploaded_ids)
    complete_resp = slices_upload_complete(item=item, uploaded_ids_map=uploaded_ids_map)
    logger.debug(f"LFS file {paths.file_path} merge {metadata.lfs_upload_part_count} uploaded slices complete response: {complete_resp}")
    verify_resp = slices_upload_verify(item=item)
    logger.debug(f"LFS file {paths.file_path} uploaded verify response: {verify_resp}")
    metadata.is_uploaded = True
    metadata.save(paths)
    logger.debug(f"LFS file {paths.file_path} - all {metadata.lfs_upload_part_count} slices uploaded successfully")

def _preupload_lfs(
    item: JOB_ITEM_T,
    status: LargeUploadStatus,
    api: CsgHubApi, 
    repo_id: str, 
    repo_type: str, 
    revision: str,
    endpoint: str,
    token: str,
):
    """Preupload LFS file and update metadata."""
    paths, metadata = item
    
    if metadata.lfs_upload_part_count is not None and metadata.lfs_upload_part_count > 0:
        return
    
    payload: Dict = {
        "operation": "upload",
        "transfers": ["basic", "multipart"],
        "objects": [
            {
                "oid": metadata.sha256,
                "size": metadata.size,
            }
        ],
        "hash_algo": "sha256",
        "upload_id": metadata.lfs_upload_id,
    }
    if revision is not None:
        payload["ref"] = {"name": unquote(revision)}  # revision has been previously 'quoted'
        
    batch_resp = api.fetch_lfs_batch_info(
        payload=payload, endpoint=endpoint, token=token,
        repo_id=repo_id, repo_type=repo_type, revision=revision, local_file=paths.file_path,
        upload_id=metadata.lfs_upload_id)
    
    objects = batch_resp.get("objects", None)
    if not isinstance(objects, list) or len(objects) < 1:
        raise ValueError(f"LFS {paths.file_path} malformed batch response objects is not list from server: {batch_resp}")
    object = objects[0]
    
    search_key = "actions"
    if not isinstance(object, dict) or search_key not in object:
        raise ValueError(f"no slices batch {search_key} info found for response of file {paths.file_path} from server: {object}")
    object_actions = object[search_key]
    
    search_key = "upload"
    if not isinstance(object_actions, dict) or search_key not in object_actions:
        raise ValueError(f"no slices batch {search_key} info found for response of file {paths.file_path} from server: {object}")
    object_upload = object_actions[search_key]
    
    search_key = "verify"
    if not isinstance(object_actions, dict) or search_key not in object_actions:
        raise ValueError(f"no slices batch {search_key} info found for response of file {paths.file_path} from server: {object}")
    object_verify = object_actions[search_key]
    
    search_key = "header"
    if not isinstance(object_upload, dict) or search_key not in object_upload:
        raise ValueError(f"no slices batch {search_key} found for response of file {paths.file_path} from server: {object}")
    object_upload_header = object_upload[search_key]
    
    href_key = "href"
    if not isinstance(object_upload, dict) or search_key not in object_upload:
        raise ValueError(f"no slices batch merge address found for response of file {paths.file_path} from server: {object}") 
    
    if not isinstance(object_upload_header, Dict):
        raise ValueError(f"incorrect lfs {paths.file_path} slices upload address from server: {object}")
    
    chunk_size = object_upload_header.pop("chunk_size")
    if chunk_size is None:
        raise ValueError(f"no chunk size found for lfs slices upload of file {paths.file_path}")
    
    total_count = len(object_upload_header)
    metadata.lfs_upload_part_count = total_count
    metadata.lfs_upload_complete_url = object_upload[href_key]
    metadata.lfs_upload_verify = object_verify
    
    sorted_keys = sorted(object_upload_header.keys(), key=lambda x: int(x))
    parsed_url = urlparse(object_upload_header.get(sorted_keys[0]))
    query_params = parse_qs(parsed_url.query)
    metadata.lfs_upload_id = query_params.get(KEY_UPLOADID, [None])[0]
    
    uploaded_ids = status.get_lfs_uploaded_slice_ids(paths.file_path)
    existing_ids = status.convert_uploaded_ids_to_map(uploaded_ids)
    for _, key in enumerate(sorted_keys):
        if existing_ids.get(key) is not None:
            continue
        upload_url = object_upload_header.get(key)
        slice_metadata = copy.deepcopy(metadata)
        slice_metadata.lfs_upload_part_count = total_count
        slice_metadata.lfs_upload_part_index = int(key)
        slice_metadata.lfs_upload_part_url = upload_url
        slice_metadata.lfs_upload_chunk_size = int(chunk_size)
        item_slice = [paths, slice_metadata]
        status.queue_uploading_lfs.put(item_slice)
    logger.debug(f"get LFS {paths.file_path} slices batch info successfully")

def _perform_lfs_slice_upload(item: JOB_ITEM_T):
    resp_header = slice_upload(item=item)
    logger.debug(f"slice upload response header: {resp_header}")
    # ('eTag', '"c681604308d0749e988746229fc16b25"')
    etag = resp_header.get("etag")
    if etag is None or etag == "":
        raise ValueError(f"invalid slice upload response header: {resp_header}, etag: {etag}")
    return etag.strip('"')

def _commit(
    items: List[JOB_ITEM_T],
    api: CsgHubApi, 
    repo_id: str, 
    repo_type: str, 
    revision: str,
    endpoint: str,
    token: str,
) -> None:
    """Commit files to the repo."""
    commit_message="Add files using upload-large-folder tool"
    payload: Dict = {
        "message": commit_message,
        "files": [
            {
                "path": paths.path_in_repo,
                "action": COMMIT_ACTION_CREATE if meta.remote_oid is None else COMMIT_ACTION_UPDATE,
                "content": meta.content_base64,
            }
            for paths, meta in items
        ]
    }
    
    commit_resp = api.create_commit(
        payload=payload, endpoint=endpoint, token=token,
        repo_id=repo_id, repo_type=repo_type, revision=revision)
    
    if commit_resp[KEY_MSG] is None or commit_resp[KEY_MSG] != MSG_OK:
        raise ValueError(f"create commit response message {commit_resp} is not {MSG_OK}")
    
    for paths, metadata in items:
        metadata.is_committed = True
        metadata.save(paths)



import transformers
from pycsghub.snapshot_download import snapshot_download
from pycsghub.utils import get_token_to_send
import os
from pathlib import Path

@classmethod
def from_pretrained(cls, pretrained_model_name_or_path,
                    *model_args, **model_kwargs):
    # first step download model
    try:
        token = get_token_to_send(None)
    except Exception:
        token = None
    if os.path.isdir(pretrained_model_name_or_path):
        path = Path(pretrained_model_name_or_path)
    else:
        path = Path(snapshot_download(pretrained_model_name_or_path, token=token))
    # second step load model
    model = cls.from_pretrained_cached(path, *model_args, **model_kwargs)
    return model



class_names = []

for i in transformers.__all__:
    try:
        if (hasattr(getattr(transformers, i), 'from_pretrained')
                and i.startswith('Auto')):
            class_names.append(i)
    except:
        pass
for i in class_names:
    try:
        locals()[i] = type(i, (), {
            'from_pretrained_cached': getattr(
                getattr(transformers, i),
                'from_pretrained'
            ),
            'from_pretrained': from_pretrained
        })
    except AttributeError as e:
        print(e)







import enum
import queue
import logging
from threading import Lock
from datetime import datetime
from typing import List, Optional, Tuple
from .local_folder import LocalUploadFileMetadata, LocalUploadFilePaths
from .consts import REPO_LFS_TYPE, REPO_REGULAR_TYPE
import base64
from io import BytesIO
from tqdm import tqdm
from .consts import META_FILE_IDENTIFIER, META_FILE_OID_PREFIX

logger = logging.getLogger(__name__)

JOB_ITEM_T = Tuple[LocalUploadFilePaths, LocalUploadFileMetadata]

class WorkerJob(enum.Enum):
    SHA256 = enum.auto()
    GET_UPLOAD_MODE = enum.auto()
    PREUPLOAD_LFS = enum.auto()
    UPLOADING_LFS = enum.auto()
    COMMIT = enum.auto()
    WAIT = enum.auto()  # if no tasks are available but we don't want to exit

class ProgressReader:
    def __init__(self, fileobj, progress_bar):
        self.fileobj = fileobj
        self.progress_bar = progress_bar
    
    def read(self, size=-1):
        data = self.fileobj.read(size)
        if data:
            self.progress_bar.update(len(data))
        return data

class LargeUploadStatus:
    """Contains information, queues and tasks for a large upload process."""

    def __init__(self, items: List[JOB_ITEM_T]):
        self.items = items
        self.queue_sha256: "queue.Queue[JOB_ITEM_T]" = queue.Queue()
        self.queue_get_upload_mode: "queue.Queue[JOB_ITEM_T]" = queue.Queue()
        self.queue_preupload_lfs: "queue.Queue[JOB_ITEM_T]" = queue.Queue()
        self.queue_uploading_lfs: "queue.Queue[JOB_ITEM_T]" = queue.Queue()
        self.queue_commit: "queue.Queue[JOB_ITEM_T]" = queue.Queue()
        self.lock = Lock()

        self.nb_workers_sha256: int = 0
        self.nb_workers_get_upload_mode: int = 0
        self.nb_workers_preupload_lfs: int = 0
        self.nb_workers_uploading_lfs: int = 0
        self.nb_workers_commit: int = 0
        self.nb_workers_waiting: int = 0
        self.last_commit_attempt: Optional[float] = None

        self._started_at = datetime.now()
        self._lfs_uploaded_ids = dict()

        # Setup queues
        num_uploaded_and_commited = 0
        for item in self.items:
            paths, metadata = item
            self._lfs_uploaded_ids[paths.file_path] = metadata.lfs_uploaded_ids
            
            if (metadata.upload_mode is not None and metadata.upload_mode == REPO_LFS_TYPE
                and metadata.is_uploaded and metadata.is_committed):
                num_uploaded_and_commited += 1
            elif (metadata.upload_mode is not None and metadata.upload_mode == REPO_REGULAR_TYPE
                  and metadata.is_committed):
                num_uploaded_and_commited += 1
            elif (metadata.sha256 is None or metadata.sha256 == ""):
                self.queue_sha256.put(item)
            elif (metadata.upload_mode is None or metadata.upload_mode == "" 
                  or metadata.remote_oid is None or metadata.remote_oid == ""):
                self.queue_get_upload_mode.put(item)
            elif (metadata.upload_mode == REPO_LFS_TYPE and not metadata.is_uploaded):
                self.queue_preupload_lfs.put(item)
            elif (not metadata.is_committed):
                self.queue_commit.put(item)
            else:
                num_uploaded_and_commited += 1
                logger.debug(f"skipping file {paths.path_in_repo} because they are already uploaded and committed")
                
        log_msg = "init upload status"
        if num_uploaded_and_commited > 0:
            log_msg = f"{log_msg}, found {len(items)} files, {num_uploaded_and_commited} of which are already uploaded and committed"
        else:
            log_msg = f"{log_msg}, found {len(items)} files"
        log_msg = f"{log_msg}, queue(sha): {self.queue_sha256.qsize()}"
        log_msg = f"{log_msg}, queue(mode): {self.queue_get_upload_mode.qsize()}"
        log_msg = f"{log_msg}, queue(preupload): {self.queue_preupload_lfs.qsize()}"
        log_msg = f"{log_msg}, queue(commit): {self.queue_commit.qsize()}"
        logger.info(log_msg)

    def current_report(self) -> str:
        """Generate a report of the current status of the large upload."""
        nb_hashed = 0
        size_hashed = 0
        nb_preuploaded = 0
        nb_lfs = 0
        nb_lfs_unsure = 0
        size_preuploaded = 0
        nb_committed = 0
        size_committed = 0
        total_size = 0
        ignored_files = 0
        total_files = 0
        nb_queued_slices = 0

        with self.lock:
            for _, metadata in self.items:
                if metadata.should_ignore:
                    ignored_files += 1
                    continue
                total_size += metadata.size
                total_files += 1
                if metadata.sha256 is not None:
                    nb_hashed += 1
                    size_hashed += metadata.size
                if metadata.upload_mode == REPO_LFS_TYPE:
                    nb_lfs += 1
                if metadata.upload_mode is None:
                    nb_lfs_unsure += 1
                if metadata.is_uploaded and metadata.upload_mode == REPO_LFS_TYPE:
                    nb_preuploaded += 1
                    size_preuploaded += metadata.size
                if metadata.is_committed:
                    nb_committed += 1
                    size_committed += metadata.size
            total_size_str = _format_size(total_size)
            nb_queued_slices = self.queue_uploading_lfs.qsize()

            now = datetime.now()
            now_str = now.strftime("%Y-%m-%d %H:%M:%S")
            elapsed = now - self._started_at
            elapsed_str = str(elapsed).split(".")[0]  # remove milliseconds

            message = "\n" + "-" * 10
            message += f" {now_str} ({elapsed_str}) "
            message += "-" * 10 + "\n"

            message += "Files:   "
            message += f"hashed {nb_hashed}/{total_files} ({_format_size(size_hashed)}/{total_size_str}) | "
            message += f"pre-uploaded: {nb_preuploaded}/{nb_lfs} ({_format_size(size_preuploaded)}/{total_size_str})"
            if nb_lfs_unsure > 0:
                message += f" (+{nb_lfs_unsure} unsure)"
            message += f" | queued-slices: {nb_queued_slices}"
            message += f" | committed: {nb_committed}/{total_files} ({_format_size(size_committed)}/{total_size_str})"
            message += f" | ignored: {ignored_files}\n"
            
            message += "Workers: "
            message += f"hashing: {self.nb_workers_sha256} | "
            message += f"get upload mode: {self.nb_workers_get_upload_mode} | "
            message += f"pre-uploading: {self.nb_workers_preupload_lfs} | "
            message += f"slices-uploading: {self.nb_workers_uploading_lfs} | "
            message += f"committing: {self.nb_workers_commit} | "
            message += f"waiting: {self.nb_workers_waiting}\n"
            message += "-" * 51

            return message

    def is_done(self) -> bool:
        with self.lock:
            return all(metadata.is_committed or metadata.should_ignore for _, metadata in self.items)

    def get_lfs_uploaded_slice_ids(self, file_path: str) -> str:
        with self.lock:
            return self._lfs_uploaded_ids.get(file_path)

    def append_lfs_uploaded_slice_id(self, file_path: str, id: int, etag: str):
        with self.lock:
            old_ids = self._lfs_uploaded_ids.get(file_path)
            id_map = self.convert_uploaded_ids_to_map(old_ids)
            new_ids = ""
            if len(id_map) == 0:
                new_ids = f"{id}:{etag}"
            else:
                if id_map.get(str(id)) is None:
                    new_ids = f"{old_ids},{id}:{etag}"
                else:
                    new_ids = old_ids
            self._lfs_uploaded_ids[file_path] = new_ids
    
    def convert_uploaded_ids_to_map(self, ids: str):
        id_map = {}
        if ids is None or ids == "":
            return id_map
        for item in ids.split(','):
            idx, etag = item.split(':', 2)
            id_map[idx] = etag
        return id_map
            
    
    def is_lfs_upload_completed(self, item: JOB_ITEM_T) -> bool:
        paths, metadata = item
        with self.lock:
            uploaded_ids = self._lfs_uploaded_ids.get(paths.file_path)
            if (uploaded_ids is not None and
                metadata.lfs_upload_id is not None and
                metadata.lfs_upload_part_count is not None and 
                len(uploaded_ids.split(",")) == metadata.lfs_upload_part_count):
                    metadata.lfs_uploaded_ids = uploaded_ids
                    return True
        return False

    def compute_file_base64(self, item: JOB_ITEM_T):
        _, meta = item
        if meta.upload_mode == REPO_LFS_TYPE:
            self._compute_lfs_file_base64(item=item)
        elif meta.upload_mode == REPO_REGULAR_TYPE:
            self._compute_regular_file_base64(item=item)
    
    def _compute_lfs_file_base64(self, item: JOB_ITEM_T):
        _, meta = item
        content = f"{META_FILE_IDENTIFIER}\n{META_FILE_OID_PREFIX}{meta.sha256}\nsize {meta.size}\n"
        content_bytes = content.encode('utf-8')
        meta.content_base64 = base64.b64encode(content_bytes).decode('utf-8')
    
    def _compute_regular_file_base64(self, item: JOB_ITEM_T):
        paths, meta = item
        # with open(paths.file_path, 'rb') as f, BytesIO() as b64_buffer:
        #     base64.encode(f, b64_buffer)
        #     meta.content_base64 = b64_buffer.getvalue().decode("utf-8")
        
        desc = f"converting {paths.file_path} to base64"
        with tqdm(total=meta.size, desc=desc, unit="B", unit_scale=True, dynamic_ncols=True) as pbar:
            with open(paths.file_path, 'rb') as f, BytesIO() as b64_buffer:
                progress_reader = ProgressReader(f, pbar)
                base64.encode(progress_reader, b64_buffer)
                b64_buffer.seek(0)
                meta.content_base64 = b64_buffer.getvalue().decode("utf-8")

def _format_size(num: int) -> str:
    """Format size in bytes into a human-readable string.

    Taken from https://stackoverflow.com/a/1094933
    """
    num_f = float(num)
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num_f) < 1000.0:
            return f"{num_f:3.1f}{unit}"
        num_f /= 1000.0
    return f"{num_f:.1f}Y"



"""Utilities to efficiently compute the SHA 256 hash of a bunch of bytes."""

from typing import BinaryIO, Optional, Tuple

from .hashlib import sha1, sha256
from tqdm import tqdm
from .status import JOB_ITEM_T

def sha_fileobj(fileobj: BinaryIO, item: JOB_ITEM_T, chunk_size: Optional[int] = None) -> Tuple[str, str]:
    """
    Computes the sha256 and sha1 hash of the given file object, by chunks of size `chunk_size`.

    Args:
        fileobj (file-like object):
            The File object to compute sha256 and sha1 for, typically obtained with `open(path, "rb")`
        chunk_size (`int`, *optional*):
            The number of bytes to read from `fileobj` at once, defaults to 1MB.

    Returns:
        `bytes`: `fileobj`'s sha256 hash as bytes
    """
    paths, meta = item
    chunk_size = chunk_size if chunk_size is not None else 1024 * 1024

    sha_256 = sha256()
    sha_1 = sha1()
    header = f'blob {meta.size}\0'.encode('utf-8')
    sha_1.update(header)
    
    desc = f"computing sha256 for {paths.file_path}"
    with tqdm(initial=0, total=meta.size, desc=desc, unit="B", unit_scale=True, dynamic_ncols=True) as pbar:
        while True:
            chunk = fileobj.read(chunk_size)
            sha_256.update(chunk)
            sha_1.update(chunk)
            pbar.update(len(chunk))
            if not chunk:
                break

    return (sha_256.digest().hex(), sha_1.hexdigest())


def git_hash(data: bytes) -> str:
    """
    Computes the git-sha1 hash of the given bytes, using the same algorithm as git.

    This is equivalent to running `git hash-object`. See https://git-scm.com/docs/git-hash-object
    for more details.

    Note: this method is valid for regular files. For LFS files, the proper git hash is supposed to be computed on the
          pointer file content, not the actual file content. However, for simplicity, we directly compare the sha256 of
          the LFS file content when we want to compare LFS files.

    Args:
        data (`bytes`):
            The data to compute the git-hash for.

    Returns:
        `str`: the git-hash of `data` as an hexadecimal string.

    Example:
    ```python
    >>> git_hash(b"Hello, World!")
    'b45ef6fec89518d314f546fd6c3025367b721684'
    ```
    """
    # Taken from https://gist.github.com/msabramo/763200
    # Note: no need to optimize by reading the file in chunks as we're not supposed to hash huge files (5MB maximum).
    sha = sha1()
    sha.update(b"blob ")
    sha.update(str(len(data)).encode())
    sha.update(b"\0")
    sha.update(data)
    return sha.hexdigest()



import logging
import time
import queue
from typing import Any, Dict, List, Optional, Tuple, TypeVar
from .status import LargeUploadStatus, WorkerJob, JOB_ITEM_T
from .consts import WAITING_TIME_IF_NO_TASKS, MAX_NB_LFS_FILES_PER_COMMIT, MAX_NB_REGULAR_FILES_PER_COMMIT

logger = logging.getLogger(__name__)

def _determine_next_job(status: LargeUploadStatus) -> Optional[Tuple[WorkerJob, List[JOB_ITEM_T]]]:
    with status.lock:
        # Commit if more than 5 minutes since last commit attempt (and at least 1 file)
        if (
            status.nb_workers_commit == 0
            and status.queue_commit.qsize() > 0
            and status.last_commit_attempt is not None
            and time.time() - status.last_commit_attempt > 5 * 60
        ):
            status.nb_workers_commit += 1
            logger.debug("job: commit (more than 5 minutes since last commit attempt)")
            return (WorkerJob.COMMIT, _get_items_to_commit(status.queue_commit))

        # Commit if at least 100 files are ready to commit
        elif status.nb_workers_commit == 0 and status.queue_commit.qsize() >= 150:
            status.nb_workers_commit += 1
            logger.debug("job: commit (>100 files ready)")
            return (WorkerJob.COMMIT, _get_items_to_commit(status.queue_commit))

        # Get upload mode if at least 10 files
        elif status.queue_get_upload_mode.qsize() >= 10:
            status.nb_workers_get_upload_mode += 1
            logger.debug("job: get upload mode (>10 files ready)")
            return (WorkerJob.GET_UPLOAD_MODE, _get_n(status.queue_get_upload_mode, 50))

        # Do uploading lfs multipart if at least 1 slice of lfs file
        elif status.queue_uploading_lfs.qsize() > 0:
            status.nb_workers_uploading_lfs += 1
            logger.debug("job: uploading lfs part")
            return (WorkerJob.UPLOADING_LFS, _get_one(status.queue_uploading_lfs))

        # Preupload LFS file if at least 1 file and no worker is preuploading LFS
        elif status.queue_preupload_lfs.qsize() > 0 and status.nb_workers_preupload_lfs == 0:
            status.nb_workers_preupload_lfs += 1
            logger.debug("job: preupload LFS (no other worker preuploading LFS)")
            return (WorkerJob.PREUPLOAD_LFS, _get_one(status.queue_preupload_lfs))
            
        # Compute sha256 if at least 1 file and no worker is computing sha256
        elif status.queue_sha256.qsize() > 0 and status.nb_workers_sha256 == 0:
            status.nb_workers_sha256 += 1
            logger.debug("job: sha256 (no other worker computing sha256)")
            return (WorkerJob.SHA256, _get_one(status.queue_sha256))

        # Get upload mode if at least 1 file and no worker is getting upload mode
        elif status.queue_get_upload_mode.qsize() > 0 and status.nb_workers_get_upload_mode == 0:
            status.nb_workers_get_upload_mode += 1
            logger.debug("job: get upload mode (no other worker getting upload mode)")
            return (WorkerJob.GET_UPLOAD_MODE, _get_n(status.queue_get_upload_mode, 50))

        # Preupload LFS file if at least 1 file
        elif status.queue_preupload_lfs.qsize() > 0 and (
            status.nb_workers_preupload_lfs == 0
        ):
            status.nb_workers_preupload_lfs += 1
            logger.debug("job: preupload LFS")
            return (WorkerJob.PREUPLOAD_LFS, _get_one(status.queue_preupload_lfs))

        # Compute sha256 if at least 1 file
        elif status.queue_sha256.qsize() > 0:
            status.nb_workers_sha256 += 1
            logger.debug("job: sha256")
            return (WorkerJob.SHA256, _get_one(status.queue_sha256))

        # Get upload mode if at least 1 file
        elif status.queue_get_upload_mode.qsize() > 0:
            status.nb_workers_get_upload_mode += 1
            logger.debug("job: get upload mode")
            return (WorkerJob.GET_UPLOAD_MODE, _get_n(status.queue_get_upload_mode, 50))

        # Commit if at least 1 file and 1 min since last commit attempt
        elif (
            status.nb_workers_commit == 0
            and status.queue_commit.qsize() > 0
            and status.last_commit_attempt is not None
            and time.time() - status.last_commit_attempt > 1 * 60
        ):
            status.nb_workers_commit += 1
            logger.debug("job: commit (1 min since last commit attempt)")
            return (WorkerJob.COMMIT, _get_items_to_commit(status.queue_commit))

        # Commit if at least 1 file all other queues are empty and all workers are waiting
        # e.g. when it's the last commit
        elif (
            status.nb_workers_commit == 0
            and status.queue_commit.qsize() > 0
            and status.queue_sha256.qsize() == 0
            and status.queue_get_upload_mode.qsize() == 0
            and status.queue_preupload_lfs.qsize() == 0
            and status.nb_workers_sha256 == 0
            and status.nb_workers_get_upload_mode == 0
            and status.nb_workers_preupload_lfs == 0
        ):
            status.nb_workers_commit += 1
            logger.debug("job: commit")
            return (WorkerJob.COMMIT, _get_items_to_commit(status.queue_commit))

        # If all queues are empty, exit
        elif all(metadata.is_committed or metadata.should_ignore for _, metadata in status.items):
            logger.info("all files have been processed! Exiting worker.")
            return None

        # If no task is available, wait
        else:
            status.nb_workers_waiting += 1
            logger.debug(f"no task available, waiting... ({WAITING_TIME_IF_NO_TASKS}s)")
            return (WorkerJob.WAIT, [])

def _get_items_to_commit(queue: "queue.Queue[JOB_ITEM_T]") -> List[JOB_ITEM_T]:
    """Special case for commit job: the number of items to commit depends on the type of files."""
    # Can take at most 50 regular files and/or 100 LFS files in a single commit
    items: List[JOB_ITEM_T] = []
    nb_lfs, nb_regular = 0, 0
    while True:
        # If empty queue => commit everything
        if queue.qsize() == 0:
            return items

        # If we have enough items => commit them
        if nb_lfs >= MAX_NB_LFS_FILES_PER_COMMIT or nb_regular >= MAX_NB_REGULAR_FILES_PER_COMMIT:
            return items

        # Else, get a new item and increase counter
        item = queue.get()
        items.append(item)
        _, metadata = item
        if metadata.upload_mode == "lfs":
            nb_lfs += 1
        else:
            nb_regular += 1

def _get_one(queue: "queue.Queue[JOB_ITEM_T]") -> List[JOB_ITEM_T]:
    return [queue.get()]

def _get_n(queue: "queue.Queue[JOB_ITEM_T]", n: int) -> List[JOB_ITEM_T]:
    return [queue.get() for _ in range(min(queue.qsize(), n))]



from typing import Literal

REPO_REGULAR_TYPE = "regular"
REPO_LFS_TYPE = "lfs"

COMMIT_ACTION_CREATE = "create"
COMMIT_ACTION_UPDATE = "update"
COMMIT_ACTION_DELETE = "delete"

KEY_MSG = "msg"
MSG_OK = "OK"
KEY_UPLOADID = "uploadId"

DEFAULT_IGNORE_PATTERNS = [
    ".git",
    ".git/*",
    "*/.git",
    "**/.git/**",
     ".cache/csghub",
    ".cache/csghub/*",
    "*/.cache/csghub",
    "**/.cache/csghub/**",
]

FORBIDDEN_FOLDERS = [".git", ".cache"]

# Timeout of aquiring file lock and logging the attempt
FILELOCK_LOG_EVERY_SECONDS = 10

WAITING_TIME_IF_NO_TASKS = 3  # seconds
MAX_NB_REGULAR_FILES_PER_COMMIT = 75
MAX_NB_LFS_FILES_PER_COMMIT = 150

META_FILE_IDENTIFIER = "version https://git-lfs.github.com/spec/v1"
META_FILE_OID_PREFIX = "oid sha256:"



import functools
import hashlib
import sys

_kwargs = {"usedforsecurity": False} if sys.version_info >= (3, 9) else {}
md5 = functools.partial(hashlib.md5, **_kwargs)
sha1 = functools.partial(hashlib.sha1, **_kwargs)
sha256 = functools.partial(hashlib.sha256, **_kwargs)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='csghub-sdk',
    version='0.7.1',
    author="opencsg",
    author_email="contact@opencsg.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(include="pycsghub*"),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "csghub-cli=pycsghub.cli:app",
        ]
    },
    install_requires=[
        "typer",
        "typing_extensions",
        "huggingface_hub>=0.22.2",
    ],
    extras_require={
        "train": [
            "torch",
            "transformers>=4.33.3",
            "datasets>=2.20.0"
        ],
    },
    python_requires=">=3.8,<3.14",
)



import logging
from typing import Dict
from pycsghub.utils import (build_csg_headers, get_endpoint, model_id_to_group_owner_name)
import requests
import base64
from pycsghub.constants import GIT_ATTRIBUTES_CONTENT, DEFAULT_REVISION, DEFAULT_LICENCE, REPO_TYPE_SPACE

logger = logging.getLogger(__name__)

class CsgHubApi:
    '''
    csghub API wrapper class
    '''

    def __init__(self):
        pass

    def fetch_upload_modes(
        self,
        payload: Dict,
        repo_id: str,
        repo_type: str,
        revision: str,
        endpoint: str,
        token: str,
    ):
        """
        Requests repo files upload modes
        """
        action_endpoint = get_endpoint(endpoint=endpoint)
        req_headers = build_csg_headers(token=token)
        fetch_url = f"{action_endpoint}/api/v1/{repo_type}s/{repo_id}/preupload/{revision}"
        response = requests.post(fetch_url, headers=req_headers, json=payload)
        if response.status_code != 200:
            logger.error(f"fetch upload modes from {fetch_url} response: {response.text}")
        response.raise_for_status()
        try:
            return response.json()
        except ValueError:
            raise ValueError(f"invalid json data for fetch upload modes from {fetch_url} response: {response.text}")

    def fetch_lfs_batch_info(
        self,
        payload: Dict,
        repo_id: str,
        repo_type: str,
        revision: str,
        endpoint: str,
        token: str,
        local_file: str,
        upload_id: str,
    ):
        """
        Requests the LFS batch endpoint to retrieve upload instructions
        """
        action_endpoint = get_endpoint(endpoint=endpoint)
        req_headers = build_csg_headers(token=token)
        batch_url = f"{action_endpoint}/{repo_type}s/{repo_id}.git/info/lfs/objects/batch"
        params = {"upload_id": upload_id}
        response = requests.post(batch_url, headers=req_headers, params=params, json=payload)
        if response.status_code != 200:
            logger.error(f"fetch LFS {local_file} batch info from {batch_url} response: {response.text}")
        response.raise_for_status()
        try:
            return response.json()
        except ValueError:
            raise ValueError(f"invalid json data for fetch LFS {local_file} batch info from {batch_url} response: {response.text}")
    
    def create_commit(
        self,
        payload: Dict,
        repo_id: str,
        repo_type: str,
        revision: str,
        endpoint: str,
        token: str,
    ):
        """
        Creates a commit in the given repo, deleting & uploading files as needed.
        """
        action_endpoint = get_endpoint(endpoint=endpoint)
        req_headers = build_csg_headers(token=token)
        commit_url = f"{action_endpoint}/api/v1/{repo_type}s/{repo_id}/commit/{revision}"
        response = requests.post(url=commit_url, headers=req_headers, json=payload)
        if response.status_code != 200:
            logger.error(f"create files commit on {commit_url} response: {response.text}")
        response.raise_for_status()
        try:
            return response.json()
        except ValueError:
            raise ValueError(f"invalid json data for create files commit on {commit_url} response: {response.text}")

    def repo_branch_exists(
        self,
        repo_id: str,
        repo_type: str,
        revision: str,
        endpoint: str,
        token: str,
    ):
        """
        Check if repo and branch exists
        """
        action_endpoint = get_endpoint(endpoint=endpoint)
        req_headers = build_csg_headers(token=token, headers={
            "Content-Type": "application/json"
        })
        action_url = f"{action_endpoint}/api/v1/{repo_type}s/{repo_id}/branches"
        response = requests.get(action_url, headers=req_headers)
        logger.debug(f"fetch {repo_type} {repo_id} branches on {action_url} response: {response.text}")
        
        if response.status_code != 200:
            return False, False
        jsonRes = response.json()
        if jsonRes["msg"] != "OK":
            return False, False

        branches = jsonRes["data"]
        for b in branches:
            if b["name"] == revision:
                return True, True
        
        return True, False
        
    def create_new_branch(
        self,
        repo_id: str,
        repo_type: str,
        revision: str,
        endpoint: str,
        token: str,
    ):
        """
        Create branch
        """
        action_endpoint = get_endpoint(endpoint=endpoint)
        req_headers = build_csg_headers(token=token, headers={
            "Content-Type": "application/json"
        })
        action_url = f"{action_endpoint}/api/v1/{repo_type}s/{repo_id}/raw/.gitattributes"
        
        GIT_ATTRIBUTES_CONTENT_BASE64 = base64.b64encode(GIT_ATTRIBUTES_CONTENT.encode()).decode()

        data = {
            "message": f"create new branch {revision}",
            "new_branch": revision,
            "content": GIT_ATTRIBUTES_CONTENT_BASE64
        }
        
        response = requests.post(action_url, json=data, headers=req_headers)
        if response.status_code != 200:
            logger.error(f"create new branch {revision} for {repo_type} {repo_id} on {action_endpoint} response: {response.text}")
        response.raise_for_status()
        try:
            return response.json()
        except ValueError:
            raise ValueError(f"invalid json data for create new branch {revision} for {repo_type} {repo_id} on {action_url} response: {response.text}")

    def create_new_repo(
        self,
        repo_id: str,
        repo_type: str,
        revision: str,
        endpoint: str,
        token: str,
    ):
        """
        Create new repo
        """        
        action_endpoint = get_endpoint(endpoint=endpoint)
        req_headers = build_csg_headers(token=token, headers={
            "Content-Type": "application/json"
        })
        action_url = f"{action_endpoint}/api/v1/{repo_type}s"
        namespace, name = model_id_to_group_owner_name(model_id=repo_id)
        data = {
            "namespace": namespace,
            "name": name,
            "nickname": name,
            "default_branch": DEFAULT_REVISION,
            "private": True,
            "license": DEFAULT_LICENCE,
        }
        
        if repo_type == REPO_TYPE_SPACE:
            resource_resp = self.get_space_resources(endpoint=endpoint)
            resources = resource_resp["data"]
            if resources is not None and len(resources) > 0:
                resource_id = resources[0]["id"]
                data["resource_id"] = resource_id
            else:
                raise ValueError(f"no any space resource found for create {repo_type} {repo_id}")
        
        response = requests.post(action_url, json=data, headers=req_headers)
        if response.status_code != 200:
            logger.error(f"create new {repo_type} {repo_id} on {action_endpoint} response: {response.text}")
        response.raise_for_status()
        try:
            return response.json()
        except ValueError:
            raise ValueError(f"invalid json data for create new {repo_type} {repo_id} on {action_url} response: {response.text}")

    def get_space_resources(
        self,
        endpoint: str,
    ):
        """
        Get space resources
        """
        action_endpoint = get_endpoint(endpoint=endpoint)
        req_headers = {
            "Content-Type": "application/json"
        }
        action_url = f"{action_endpoint}/api/v1/space_resources"
        params = {"deploy_type": "0"}
        response = requests.get(action_url, params=params, headers=req_headers)
        if response.status_code != 200:
            logger.error(f"query space resources on {action_endpoint} response: {response.text}")
        response.raise_for_status()
        try:
            return response.json()
        except ValueError:
            raise ValueError(f"invalid json data for query space resources on {action_url} response: {response.text}")



from fnmatch import fnmatch
from pathlib import Path
from typing import Callable, Generator, Iterable, List, Optional, TypeVar, Union

T = TypeVar("T")

def filter_repo_objects(
    items: Iterable[T],
    *,
    allow_patterns: Optional[Union[List[str], str]] = None,
    ignore_patterns: Optional[Union[List[str], str]] = None,
    key: Optional[Callable[[T], str]] = None,
) -> Generator[T, None, None]:
    if isinstance(allow_patterns, str):
        allow_patterns = [allow_patterns]

    if isinstance(ignore_patterns, str):
        ignore_patterns = [ignore_patterns]

    if allow_patterns is not None:
        allow_patterns = [_add_wildcard_to_directories(p) for p in allow_patterns]
    if ignore_patterns is not None:
        ignore_patterns = [_add_wildcard_to_directories(p) for p in ignore_patterns]

    if key is None:

        def _identity(item: T) -> str:
            if isinstance(item, str):
                return item
            if isinstance(item, Path):
                return str(item)
            raise ValueError(f"please provide `key` argument in `filter_repo_objects`: `{item}` is not a string.")

        key = _identity

    for item in items:
        path = key(item)

        if allow_patterns is not None and not any(fnmatch(path, r) for r in allow_patterns):
            continue

        if ignore_patterns is not None and any(fnmatch(path, r) for r in ignore_patterns):
            continue

        yield item


def _add_wildcard_to_directories(pattern: str) -> str:
    if pattern[-1] == "/":
        return pattern + "*"
    return pattern



import tempfile
from functools import partial
from http.cookiejar import CookieJar
from pathlib import Path
from typing import Optional, Union, List, Dict
import requests
from huggingface_hub.utils import filter_repo_objects
from requests.adapters import Retry
from tqdm import tqdm
from pycsghub import utils
from pycsghub.cache import ModelFileSystemCache
from pycsghub.utils import (build_csg_headers,
                            get_cache_dir,
                            model_id_to_group_owner_name,
                            pack_repo_file_info,
                            get_file_download_url,
                            get_endpoint)
from pycsghub.constants import (API_FILE_DOWNLOAD_RETRY_TIMES,
                                API_FILE_DOWNLOAD_TIMEOUT,
                                API_FILE_DOWNLOAD_CHUNK_SIZE,
                                DEFAULT_REVISION)
from pycsghub.errors import FileDownloadError
import os
from pycsghub.errors import InvalidParameter


def try_to_load_from_cache():
    pass


def cached_download():
    pass


def csg_hub_download():
    pass


def get_csg_hub_url():
    pass


def file_download(
        repo_id: str,
        *,
        file_name: str = None,
        revision: Optional[str] = DEFAULT_REVISION,
        cache_dir: Union[str, Path, None] = None,
        local_files_only: Optional[bool] = False,
        cookies: Optional[CookieJar] = None,
        allow_patterns: Optional[Union[List[str], str]] = None,
        ignore_patterns: Optional[Union[List[str], str]] = None,
        headers: Optional[Dict[str, str]] = None,
        endpoint: Optional[str] = None,
        token: Optional[str] = None,
        repo_type: Optional[str] = None,
        source: Optional[str] = None,
) -> str:
    if cache_dir is None:
        cache_dir = get_cache_dir(repo_type=repo_type)
    if isinstance(cache_dir, Path):
        cache_dir = str(cache_dir)
    temporary_cache_dir = os.path.join(cache_dir, 'temp')
    os.makedirs(temporary_cache_dir, exist_ok=True)

    if file_name is None:
        raise InvalidParameter('file_name cannot be None, if you want '
                               'to load single file from repo {}'.format(repo_id))

    group_or_owner, name = model_id_to_group_owner_name(repo_id)
    name = name.replace('.', '___')

    cache = ModelFileSystemCache(cache_dir, group_or_owner, name)

    if local_files_only:
        if len(cache.cached_files) == 0:
            raise ValueError(
                'Cannot find the requested files in the cached path and outgoing'
                ' traffic has been disabled. To enable model look-ups and downloads'
                " online, set 'local_files_only' to False.")
        return cache.get_root_location()
    else:
        download_endpoint = get_endpoint(endpoint=endpoint)
        # make headers
        # todo need to add cookies？
        repo_info = utils.get_repo_info(repo_id=repo_id,
                                        revision=revision,
                                        token=token,
                                        endpoint=download_endpoint,
                                        repo_type=repo_type,
                                        source=source)

        assert repo_info.sha is not None, "Repo info returned from server must have a revision sha."
        assert repo_info.siblings is not None, "Repo info returned from server must have a siblings list."
        model_files = list(
            filter_repo_objects(
                items=[f.rfilename for f in repo_info.siblings],
                allow_patterns=allow_patterns,
                ignore_patterns=ignore_patterns,
            )
        )
        if file_name not in model_files:
            raise InvalidParameter('file {} not in repo {}'.format(file_name, repo_id))

        with tempfile.TemporaryDirectory(
                dir=temporary_cache_dir) as temp_cache_dir:
            repo_file_info = pack_repo_file_info(file_name, revision)
            if not cache.exists(repo_file_info):
                file_name = os.path.basename(repo_file_info['Path'])
                # get download url
                url = get_file_download_url(
                    model_id=repo_id,
                    file_path=file_name,
                    revision=revision,
                    endpoint=download_endpoint,
                    repo_type=repo_type,
                    source=source)
                # todo support parallel download api
                http_get(
                    url=url,
                    local_dir=temp_cache_dir,
                    file_name=file_name,
                    headers=headers,
                    cookies=cookies,
                    token=token)

                # todo using hash to check file integrity
                temp_file = os.path.join(temp_cache_dir, file_name)
                cache.put_file(repo_file_info, temp_file)
            else:
                print(
                    f'File {file_name} already in cache {cache.get_root_location()}, skip downloading!'
                )
        cache.save_model_version(revision_info={'Revision': revision})
        return os.path.join(cache.get_root_location(), file_name)


def http_get(*,
             url: str,
             local_dir: str,
             file_name: str,
             headers: dict = None,
             cookies: CookieJar = None,
             token: str = None) -> None:
    '''
    download core API，using python request to download file to local cache dirs
    :param token: csghub token
    :param url: url to download
    :param local_dir: local dir to download
    :param file_name: file name to download
    :param headers: http headers
    :param cookies: http cookies
    :return: None
    '''
    tempfile_mgr = partial(tempfile.NamedTemporaryFile, mode='wb', dir=local_dir, delete=False)
    get_headers = build_csg_headers(token=token, headers=headers)
    total_content_length = 0
    with tempfile_mgr() as temp_file:
        # retry sleep 0.5s, 1s, 2s, 4s
        retry = Retry(total=API_FILE_DOWNLOAD_RETRY_TIMES, backoff_factor=1, allowed_methods=['GET'])
        while True:
            try:
                downloaded_size = temp_file.tell()
                if downloaded_size > 0:
                    get_headers['Range'] = 'bytes=%d-' % downloaded_size
                r = requests.get(url, headers=get_headers, stream=True,
                                 cookies=cookies, timeout=API_FILE_DOWNLOAD_TIMEOUT)
                r.raise_for_status()
                accept_ranges = r.headers.get('Accept-Ranges')
                content_length = r.headers.get('Content-Length')
                if accept_ranges == 'bytes':
                    if downloaded_size == 0:
                        total_content_length = int(content_length) if content_length is not None else None
                else:
                    if downloaded_size > 0:
                        temp_file.truncate(0)
                        downloaded_size = temp_file.tell()
                    total_content_length = int(content_length) if content_length is not None else None

                progress = tqdm(
                    unit='B',
                    unit_scale=True,
                    unit_divisor=1024,
                    total=total_content_length,
                    initial=downloaded_size,
                    desc="Downloading {}".format(file_name),
                )
                for chunk in r.iter_content(chunk_size=API_FILE_DOWNLOAD_CHUNK_SIZE):
                    if chunk:
                        progress.update(len(chunk))
                        temp_file.write(chunk)
                progress.close()
                break
            except Exception as e:
                retry = retry.increment('GET', url, error=e)
                retry.sleep()

    downloaded_length = os.path.getsize(temp_file.name)
    if total_content_length != downloaded_length:
        os.remove(temp_file.name)
        msg = 'File %s download incomplete, content_length: %s but the file downloaded length: %s, please download again' % (
            file_name, total_content_length, downloaded_length)
        raise FileDownloadError(msg)
    # fix folder recursive issue
    os.makedirs(os.path.dirname(os.path.join(local_dir, file_name)), exist_ok=True)
    os.replace(temp_file.name, os.path.join(local_dir, file_name))
    return


if __name__ == '__main__':
    token = "your_access_token"

    url = "https://hub.opencsg.com/api/v1/models/wayne0019/lwfmodel/resolve/lfsfile.bin"
    local_dir = '/home/test/'
    file_name = 'test.txt'
    headers = None
    cookies = None
    http_get(url=url,
             token=token,
             local_dir=local_dir,
             file_name=file_name,
             headers=headers,
             cookies=cookies)



import typer
import os
import logging
from typing import List, Optional
from typing_extensions import Annotated
from pycsghub.cmd import repo, inference, finetune
from pycsghub.cmd.repo_types import RepoType
from importlib.metadata import version
from pycsghub.constants import DEFAULT_CSGHUB_DOMAIN, DEFAULT_REVISION
from .upload_large_folder.main import upload_large_folder_internal
from pycsghub.constants import REPO_SOURCE_CSG

logger = logging.getLogger(__name__)

app = typer.Typer(
    add_completion=False,
    no_args_is_help=True,
)

def version_callback(value: bool):
    if value:
        pkg_version = version("csghub-sdk")
        print(f"csghub-cli version {pkg_version}")
        raise typer.Exit()

OPTIONS = {
    "repoID": typer.Argument(help="The ID of the repo. (e.g. `username/repo-name`)."),
    "localPath": typer.Argument(help="Local path to the file to upload. Defaults to the relative path of the file of repo of OpenCSG Hub."),
    "pathInRepo": typer.Argument(help="Path of the folder in the repo. Defaults to the relative path of the file or folder."),
    "repoType": typer.Option("-t", "--repo-type", help="Specify the repository type."),
    "revision": typer.Option("-r", "--revision", help="An optional Git revision id which can be a branch name"),
    "cache_dir": typer.Option("-cd", "--cache-dir", help="Path to the directory where to save the downloaded files."),
    "endpoint": typer.Option("-e", "--endpoint", help="The address of the request to be sent."),
    "username": typer.Option("-u", "--username", help="Logon account of OpenCSG Hub."),
    "token": typer.Option("-k", "--token", help="A User Access Token generated from https://opencsg.com/settings/access-token"),
    "allow_patterns": typer.Option("--allow-patterns", help="Allow patterns for files to be downloaded."),
    "ignore_patterns": typer.Option("--ignore-patterns", help="Ignore patterns for files to be downloaded."),
    "version": typer.Option(None, "-V", "--version", callback=version_callback, is_eager=True, help="Show the version and exit."),
    "limit": typer.Option("--limit", help="Number of items to list"),
    "localFolder": typer.Argument(help="Local path to the folder to upload."),
    "num_workers": typer.Option("-n","--num-workers", help="Number of concurrent upload workers."),
    "print_report": typer.Option("--print-report", help="Whether to print a report of the upload progress. Defaults to True."),
    "print_report_every": typer.Option("--print-report-every", help="Frequency at which the report is printed. Defaults to 60 seconds."),
    "log_level": typer.Option("INFO", "-L", "--log-level",
        help="set log level (DEBUG, INFO, WARNING, ERROR, CRITICAL)",
        case_sensitive=False,
    ),
    "source": typer.Option("--source", help="Specify the source of the repository (e.g. 'csg', 'hf', 'ms')."),
}

@app.command(name="download", help="Download model/dataset from OpenCSG Hub", no_args_is_help=True)
def download(
        repo_id: Annotated[str, OPTIONS["repoID"]],
        repo_type: Annotated[RepoType, OPTIONS["repoType"]] = RepoType.MODEL, 
        revision: Annotated[Optional[str], OPTIONS["revision"]] = DEFAULT_REVISION,
        endpoint: Annotated[Optional[str], OPTIONS["endpoint"]] = DEFAULT_CSGHUB_DOMAIN,
        token: Annotated[Optional[str], OPTIONS["token"]] = None,
        cache_dir: Annotated[Optional[str], OPTIONS["cache_dir"]] = None,
        allow_patterns: Annotated[Optional[List[str]], OPTIONS["allow_patterns"]] = None,
        ignore_patterns: Annotated[Optional[List[str]], OPTIONS["ignore_patterns"]] = None,
        source: Annotated[str, OPTIONS["source"]] = REPO_SOURCE_CSG,
    ):
    repo.download(
        repo_id=repo_id,
        repo_type=repo_type.value,
        revision=revision,
        cache_dir=cache_dir,
        endpoint=endpoint,
        token=token,
        allow_patterns=allow_patterns,
        ignore_patterns=ignore_patterns,
        source=source,
    )

@app.command(name="upload", help="Upload repository files to OpenCSG Hub", no_args_is_help=True)
def upload(
        repo_id: Annotated[str, OPTIONS["repoID"]],
        local_path: Annotated[str, OPTIONS["localPath"]],
        path_in_repo: Annotated[str, OPTIONS["pathInRepo"]] = "",
        repo_type: Annotated[RepoType, OPTIONS["repoType"]] = RepoType.MODEL,
        revision: Annotated[Optional[str], OPTIONS["revision"]] = DEFAULT_REVISION,
        endpoint: Annotated[Optional[str], OPTIONS["endpoint"]] = DEFAULT_CSGHUB_DOMAIN,
        token: Annotated[Optional[str], OPTIONS["token"]] = None,
        user_name: Annotated[Optional[str], OPTIONS["username"]] = "",
    ):
    # File upload
    if os.path.isfile(local_path):
        repo.upload_files(
            repo_id=repo_id, 
            repo_type=repo_type.value, 
            repo_file=local_path,
            path_in_repo=path_in_repo,
            revision=revision, 
            endpoint=endpoint,
            token=token
        )
    # Folder upload
    else:
        repo.upload_folder(
            repo_id=repo_id, 
            repo_type=repo_type.value,
            local_path=local_path,
            path_in_repo=path_in_repo,
            revision=revision,
            endpoint=endpoint,
            token=token,
            user_name=user_name
        )
        
@app.command(name="upload-large-folder", help="Upload large folder to OpenCSG Hub using multiple workers", no_args_is_help=True)
def upload_large_folder(
    repo_id: Annotated[str, OPTIONS["repoID"]],
    local_path: Annotated[str, OPTIONS["localFolder"]],
    repo_type: Annotated[RepoType, OPTIONS["repoType"]] = RepoType.MODEL,
    revision: Annotated[Optional[str], OPTIONS["revision"]] = DEFAULT_REVISION,
    endpoint: Annotated[Optional[str], OPTIONS["endpoint"]] = DEFAULT_CSGHUB_DOMAIN,
    token: Annotated[Optional[str], OPTIONS["token"]] = None,
    allow_patterns: Annotated[Optional[List[str]], OPTIONS["allow_patterns"]] = None,
    ignore_patterns: Annotated[Optional[List[str]], OPTIONS["ignore_patterns"]] = None,
    num_workers: Annotated[int, OPTIONS["num_workers"]] = None,
    print_report: Annotated[bool, OPTIONS["print_report"]] = False,
    print_report_every: Annotated[int, OPTIONS["print_report_every"]] = 60,
):
    upload_large_folder_internal(
        repo_id=repo_id,
        local_path=local_path,
        repo_type=repo_type.value,
        revision=revision,
        endpoint=endpoint,
        token=token,
        allow_patterns=allow_patterns,
        ignore_patterns=ignore_patterns,
        num_workers=num_workers,
        print_report=print_report,
        print_report_every=print_report_every,
    )


inference_app = typer.Typer(
    no_args_is_help=True,
    help="Manage inference instances on OpenCSG Hub"
)
app.add_typer(inference_app, name="inference")

@inference_app.command(name="list", help="List inference instances", no_args_is_help=True)
def list_inferences(
    user_name: Annotated[str, OPTIONS["username"]],
    token: Annotated[str, OPTIONS["token"]] = None,
    endpoint: Annotated[Optional[str], OPTIONS["endpoint"]] = DEFAULT_CSGHUB_DOMAIN,
    limit: Annotated[Optional[int], OPTIONS["limit"]] = 50,
):
    inference.list(
        user_name=user_name,
        token=token,
        endpoint=endpoint,
        limit=limit,
    )

@inference_app.command(name="start", help="Start inference instance", no_args_is_help=True)
def start_inference(
    model: str = typer.Argument(..., help="model to use for inference"),
    id: int = typer.Argument(..., help="ID of the inference instance to start"),
    token: Annotated[Optional[str], OPTIONS["token"]] = None,
    endpoint: Annotated[Optional[str], OPTIONS["endpoint"]] = DEFAULT_CSGHUB_DOMAIN,
):
    inference.start(
        id=id,
        model=model,
        token=token,
        endpoint=endpoint,
    )


@inference_app.command(name="stop", help="Stop inference instance", no_args_is_help=True)
def stop_inference(
    model: str = typer.Argument(..., help="model to use for inference"),
    id: int = typer.Argument(..., help="ID of the inference instance to start"),
    token: Annotated[Optional[str], OPTIONS["token"]] = None,
    endpoint: Annotated[Optional[str], OPTIONS["endpoint"]] = DEFAULT_CSGHUB_DOMAIN,
):
    inference.stop(
        id=id,
        model=model,
        token=token,
        endpoint=endpoint,
    )
 
finetune_app = typer.Typer(
    no_args_is_help=True,
    help="Manage fine-tuning instances on OpenCSG Hub"
)
app.add_typer(finetune_app, name="finetune")

@finetune_app.command(name="list", help="List fine-tuning instances", no_args_is_help=True)
def list_finetune(
    user_name: Annotated[str, OPTIONS["username"]],
    token: Annotated[str, OPTIONS["token"]] = None,
    endpoint: Annotated[Optional[str], OPTIONS["endpoint"]] = DEFAULT_CSGHUB_DOMAIN,
    limit: Annotated[Optional[int], OPTIONS["limit"]] = 50,
):
    finetune.list(
        user_name=user_name,
        token=token,
        endpoint=endpoint,
        limit=limit,
    )

@finetune_app.command(name="start", help="Start fine-tuning instance", no_args_is_help=True)
def start_finetune(
    model: str = typer.Argument(..., help="model to use for fine-tuning"),
    id: int = typer.Argument(..., help="ID of the fine-tuning instance to start"),
    token: Annotated[Optional[str], OPTIONS["token"]] = None,
    endpoint: Annotated[Optional[str], OPTIONS["endpoint"]] = DEFAULT_CSGHUB_DOMAIN,
):
    finetune.start(
        id=id,
        model=model,
        token=token,
        endpoint=endpoint,
    )

@finetune_app.command(name="stop", help="Stop fine-tuning instance", no_args_is_help=True)
def stop_finetune(
    model: str = typer.Argument(..., help="model to use for fine-tuning"),
    id: int = typer.Argument(..., help="ID of the fine-tuning instance to stop"),
    token: Annotated[Optional[str], OPTIONS["token"]] = None,
    endpoint: Annotated[Optional[str], OPTIONS["endpoint"]] = DEFAULT_CSGHUB_DOMAIN,
):
    finetune.stop(
        id=id,
        model=model,
        token=token,
        endpoint=endpoint,
    )

@app.callback(
    invoke_without_command=True,
    no_args_is_help=True,
    help="OpenCSG Hub CLI",
    context_settings={
        "help_option_names": ["-h", "--help"],
    }
)
def main(
    version: bool = OPTIONS["version"],
    log_level: str = OPTIONS["log_level"]
):
    # for example: format='%(asctime)s - %(name)s:%(funcName)s:%(lineno)d - %(levelname)s - %(message)s',
    logging.basicConfig(
        level=getattr(logging, log_level),
        format='%(asctime)s - %(levelname)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S',
        handlers=[logging.StreamHandler()]
    )
    pass


if __name__ == "__main__":
    app()
